﻿Option Explicit On
Option Strict On

Imports System.ComponentModel
Imports System.Globalization
Imports System.IO
Imports System.Runtime.InteropServices
Imports System.Security.Permissions
Imports System.Transactions

Imports MvpsOrg.Dotnet.Win32.NativeMethods
Imports MvpsOrg.Dotnet.Win32.SafeHandles

Namespace MvpsOrg.Dotnet.Win32.Transactions.IO
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class Directory
        Private Sub New()
            '
        End Sub

#Region "Create"
#Region "Not transacted"
        Public Shared Sub Create( _
            ByVal Path As String, _
            ByVal TemplateDirectoryName As String _
        )
            Dim Success As Boolean = _
                CreateDirectoryEx( _
                    TemplateDirectoryName, _
                    Path, _
                    New SECURITY_ATTRIBUTES() _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub

        Public Shared Sub Create( _
            ByVal DirectoryName As String _
        )
            Create( _
                DirectoryName, _
                Nothing _
            )
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CreateTransactedKtm( _
            ByVal DirectoryName As String _
        )
            CreateTransactedKtm( _
                DirectoryName, _
                Nothing _
            )
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            CreateTransacted( _
                DirectoryName, _
                Nothing, _
                Transaction _
            )
        End Sub

        Public Shared Sub CreateTransactedKtm( _
            ByVal DirectoryName As String, _
            ByVal TemplateDirectoryName As String _
        )
            Dim CurrentTransaction As Ktm.Transaction = _
                Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CreateTransacted( _
                DirectoryName, _
                TemplateDirectoryName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String, _
            ByVal TemplateDirectoryName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                CreateTransactedCore( _
                    DirectoryName, _
                    TemplateDirectoryName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String _
        )
            CreateTransacted( _
                DirectoryName, _
                DirectCast(Nothing, Ktm.Transaction) _
            )
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String, _
            ByVal Transaction As Transaction _
        )
            CreateTransacted( _
                DirectoryName, _
                Nothing, _
                Transaction _
            )
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String, _
            ByVal TemplateDirectoryName As String _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(CurrentTransaction)
                CreateTransactedCore( _
                    DirectoryName, _
                    TemplateDirectoryName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CreateTransacted( _
            ByVal DirectoryName As String, _
            ByVal TemplateDirectoryName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                CreateTransactedCore( _
                    DirectoryName, _
                    TemplateDirectoryName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub CreateTransactedCore( _
            ByVal DirectoryName As String, _
            ByVal TemplateDirectoryName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                CreateDirectoryTransacted( _
                    TemplateDirectoryName, _
                    DirectoryName, _
                    New SECURITY_ATTRIBUTES(), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "Delete"
#Region "Not transacted"
        Public Shared Sub Delete( _
            ByVal DirectoryName As String _
        )
            Dim Success As Boolean = _
                RemoveDirectory(DirectoryName)
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub DeleteTransactedKtm( _
            ByVal DirectoryName As String _
        )
            Dim CurrenTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrenTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            DeleteTransacted( _
                DirectoryName, _
                CurrenTransaction _
            )
        End Sub

        Public Shared Sub DeleteTransacted( _
            ByVal DirectoryName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                DeleteTransactedCore( _
                    DirectoryName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub DeleteTransacted( _
             ByVal DirectoryName As String _
         )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            DeleteTransacted(DirectoryName, CurrentTransaction)
        End Sub

        Public Shared Sub DeleteTransacted( _
            ByVal DirectoryName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                DeleteTransactedCore( _
                    DirectoryName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub DeleteTransactedCore( _
            ByVal DirectoryName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                RemoveDirectoryTransacted( _
                    DirectoryName, _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "GetFiles"
#Region "Not transacted"
        Public Shared Function GetFiles( _
            ByVal FileName As String _
        ) As String()
            Return GetFiles(FileName, False)
        End Function

        Public Shared Function GetFiles( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean _
        ) As String()
            Dim FoundFiles As New List(Of String)()
            Dim FindData As New WIN32_FIND_DATA()
            Using FindFileHandle As SafeFindHandle = _
                FindFirstFileEx( _
                    FileName, _
                    FINDEX_INFO_LEVELS.FindExInfoStandard, _
                    FindData, _
                    FINDEX_SEARCH_OPS.FindExSearchNameMatch, _
                    IntPtr.Zero, _
                    If(CaseSensitive, AdditionalFlags_Flags.FIND_FIRST_EX_CASE_SENSITIVE, AdditionalFlags_Flags.None) _
                )

                If FindFileHandle.IsInvalid Then
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
                Dim Success As Boolean
                Do
                    FoundFiles.Add(String.Copy(FindData.cFileName))
                    Success = FindNextFile(FindFileHandle, FindData)
                Loop While Success
                Return FoundFiles.ToArray()
            End Using
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetFilesTransactedKtm( _
            ByVal FileName As String _
        ) As String()
            Return _
                GetFilesTransactedKtm( _
                    FileName, _
                    False _
                )
        End Function

        Public Shared Function GetFilesTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As String()
            Return _
                GetFilesTransacted( _
                    FileName, _
                    False, _
                    Transaction _
                )
        End Function

        Public Shared Function GetFilesTransactedKtm( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean _
        ) As String()
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetFilesTransacted( _
                    FileName, _
                    CaseSensitive, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetFilesTransacted( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean, _
            ByVal Transaction As Ktm.Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetFilesTransactedCore( _
                        FileName, _
                        CaseSensitive, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetFilesTransacted( _
            ByVal FileName As String _
        ) As String()
            Return _
                GetFilesTransacted( _
                    FileName, _
                    False _
                )
        End Function

        Public Shared Function GetFilesTransacted( _
                    ByVal FileName As String, _
                    ByVal Transaction As Transaction _
                ) As String()
            Return _
                GetFilesTransacted( _
                    FileName, _
                    False, _
                    Transaction _
                )
        End Function

        Public Shared Function GetFilesTransacted( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean _
        ) As String()
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetFilesTransacted( _
                    FileName, _
                    CaseSensitive, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetFilesTransacted( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean, _
            ByVal Transaction As Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetFilesTransactedCore( _
                        FileName, _
                        CaseSensitive, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetFilesTransactedCore( _
            ByVal FileName As String, _
            ByVal CaseSensitive As Boolean, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As String()
            Dim FoundFiles As New List(Of String)()
            Dim FindData As New WIN32_FIND_DATA()
            Using FindFileHandle As SafeFindHandle = _
                FindFirstFileTransacted( _
                    FileName, _
                    FINDEX_INFO_LEVELS.FindExInfoStandard, _
                    FindData, _
                    FINDEX_SEARCH_OPS.FindExSearchNameMatch, _
                    IntPtr.Zero, _
                    If(CaseSensitive, AdditionalFlags_Flags.FIND_FIRST_EX_CASE_SENSITIVE, AdditionalFlags_Flags.None), _
                    TransactionHandle _
                )

                If FindFileHandle.IsInvalid Then
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
                Dim Success As Boolean
                Do
                    FoundFiles.Add(String.Copy(FindData.cFileName))
                    Success = FindNextFile(FindFileHandle, FindData)
                Loop While Success
                Return FoundFiles.ToArray()
            End Using
        End Function
#End Region
#End Region

#Region "GetHardLinks"
#Region "Not transacted"
        Public Shared Function GetHardLinks( _
            ByVal FileName As String _
        ) As String()
            Dim FoundHardLinks As New List(Of String)()
            Const BufferLength As Integer = MAX_PATH  ' Chosen arbitrarily.
            Dim HardLink As String = Space(BufferLength)
            Dim HardLinkLength As Integer = HardLink.Length
            Dim FindFileNameHandle As SafeFindHandle = _
                FindFirstFileNameW( _
                    FileName, _
                    0, _
                    HardLinkLength, _
                    HardLink _
                )
            If FindFileNameHandle.IsInvalid Then
                If Marshal.GetLastWin32Error() = ERROR_MORE_DATA Then
                    HardLink = Space(HardLinkLength)

                    ' The find handle has been invalid.
                    FindFileNameHandle = _
                        FindFirstFileNameW( _
                            FileName, _
                            0, _
                            HardLinkLength, _
                            HardLink _
                        )
                    If FindFileNameHandle.IsInvalid Then
                        Dim ex As New Win32Exception()
                        Throw New IOException(ex.Message, ex)
                    End If
                Else
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
            End If
            Using FindFileNameHandle
                Dim Success As Boolean
                Do
                    FoundHardLinks.Add(Left(HardLink, HardLinkLength))
                    HardLink = Space(BufferLength)
                    HardLinkLength = HardLink.Length
                    Success = _
                        FindNextFileNameW( _
                            FindFileNameHandle, _
                            HardLinkLength, _
                            HardLink _
                        )
                    If Not Success Then

                        ' No hardlinks found.
                        If Marshal.GetLastWin32Error() = ERROR_HANDLE_EOF Then
                            Exit Do
                        ElseIf Marshal.GetLastWin32Error() = ERROR_MORE_DATA Then
                            HardLink = Space(HardLinkLength)
                            Success = _
                                FindNextFileNameW( _
                                    FindFileNameHandle, _
                                    HardLinkLength, _
                                    HardLink _
                                )
                            If Not Success Then
                                Dim ex As New Win32Exception()
                                Throw New IOException(ex.Message, ex)
                            End If
                        Else
                            Dim ex As New Win32Exception()
                            Throw New IOException(ex.Message, ex)
                        End If
                    End If
                Loop While Success
                Return FoundHardLinks.ToArray()
            End Using
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetHardLinksTransactedKtm( _
            ByVal FileName As String _
        ) As String()
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetHardLinksTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetHardLinksTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetHardLinksTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetHardLinksTransacted( _
            ByVal FileName As String _
        ) As String()
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetHardLinksTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetHardLinksTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetHardLinksTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetHardLinksTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As String()
            Dim FoundHardLinks As New List(Of String)()
            Const BufferLength As Integer = MAX_PATH  ' Chosen arbitrarily.
            Dim HardLink As String = Space(BufferLength)
            Dim HardLinkLength As Integer = HardLink.Length
            Dim FindFileNameHandle As SafeFindHandle = _
                FindFirstFileNameTransactedW( _
                    FileName, _
                    0, _
                    HardLinkLength, _
                    HardLink, _
                    TransactionHandle _
                )
            If FindFileNameHandle.IsInvalid Then
                If Marshal.GetLastWin32Error() = ERROR_MORE_DATA Then
                    HardLink = Space(HardLinkLength)

                    ' The find handle has been invalid.
                    FindFileNameHandle = _
                        FindFirstFileNameTransactedW( _
                            FileName, _
                            0, _
                            HardLinkLength, _
                            HardLink, _
                            TransactionHandle _
                        )
                    If FindFileNameHandle.IsInvalid Then
                        Dim ex As New Win32Exception()
                        Throw New IOException(ex.Message, ex)
                    End If
                Else
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
            End If
            Using FindFileNameHandle
                Dim Success As Boolean
                Do
                    FoundHardLinks.Add(Left(HardLink, HardLinkLength))
                    HardLink = Space(BufferLength)
                    HardLinkLength = HardLink.Length
                    Success = _
                        FindNextFileNameW( _
                            FindFileNameHandle, _
                            HardLinkLength, _
                            HardLink _
                        )
                    If Not Success Then

                        ' No hard links found.
                        If Marshal.GetLastWin32Error() = ERROR_HANDLE_EOF Then
                            Exit Do
                        ElseIf Marshal.GetLastWin32Error() = ERROR_MORE_DATA Then
                            HardLink = Space(HardLinkLength)
                            Success = _
                                FindNextFileNameW( _
                                    FindFileNameHandle, _
                                    HardLinkLength, _
                                    HardLink _
                                )
                            If Not Success Then
                                Dim ex As New Win32Exception()
                                Throw New IOException(ex.Message, ex)
                            End If
                        Else
                            Dim ex As New Win32Exception()
                            Throw New IOException(ex.Message, ex)
                        End If
                    End If
                Loop While Success
                Return FoundHardLinks.ToArray()
            End Using
        End Function
#End Region
#End Region

#Region "GetStreams"
#Region "Not transacted"
        Public Shared Function GetStreams( _
            ByVal FileName As String _
        ) As String()
            Dim FoundStreams As New List(Of String)()
            Dim FindData As New WIN32_FIND_STREAM_DATA()
            Using FindStreamHandle As SafeFindHandle = _
                FindFirstStreamW( _
                    FileName, _
                    STREAM_INFO_LEVELS.FindStreamInfoStandard, _
                    FindData, _
                    0 _
                )

                If FindStreamHandle.IsInvalid Then
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
                Dim Success As Boolean
                Do
                    FoundStreams.Add(String.Copy(FindData.cStreamName))
                    Success = FindNextStreamW(FindStreamHandle, FindData)
                Loop While Success
                Return FoundStreams.ToArray()
            End Using
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetStreamsTransactedKtm( _
            ByVal FileName As String _
        ) As String()
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetStreamsTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetStreamsTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.Current.GetHandle()
                Return _
                    GetStreamsTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetStreamsTransacted( _
            ByVal FileName As String _
        ) As String()
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetStreamsTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetStreamsTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As String()
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction.Current)
                Return _
                    GetStreamsTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetStreamsTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As String()
            Dim FoundStreams As New List(Of String)()
            Dim FindData As New WIN32_FIND_STREAM_DATA()
            Using FindStreamHandle As SafeFindHandle = _
                FindFirstStreamTransactedW( _
                    FileName, _
                    STREAM_INFO_LEVELS.FindStreamInfoStandard, _
                    FindData, _
                    0, _
                    TransactionHandle _
                )

                If FindStreamHandle.IsInvalid Then
                    Dim ex As New Win32Exception()
                    Throw New IOException(ex.Message, ex)
                End If
                Dim Success As Boolean
                Do
                    FoundStreams.Add(String.Copy(FindData.cStreamName))
                    Success = FindNextStreamW(FindStreamHandle, FindData)
                Loop While Success
                Return FoundStreams.ToArray()
            End Using
        End Function
#End Region
#End Region

#Region "CreateSymbolicLink"
#Region "Not transacted"
        Public Shared Sub CreateSymbolicLink( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetDirectoryName As String _
        )
            Dim Success As Boolean = _
                NativeMethods.CreateSymbolicLink( _
                    SymbolicLinkFileName, _
                    TargetDirectoryName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_DIRECTORY _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CreateSymbolicLinkTransactedKtm( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetDirectoryName As String _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CreateSymbolicLinkTransacted( _
                SymbolicLinkFileName, _
                TargetDirectoryName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetDirectoryName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                File.CreateSymbolicLinkTransactedCore( _
                    SymbolicLinkFileName, _
                    TargetDirectoryName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_DIRECTORY, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetDirectoryName As String _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            CreateSymbolicLinkTransacted( _
                SymbolicLinkFileName, _
                TargetDirectoryName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetDirectoryName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                File.CreateSymbolicLinkTransactedCore( _
                    SymbolicLinkFileName, _
                    TargetDirectoryName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_DIRECTORY, _
                    TransactionHandle _
                )
            End Using
        End Sub
#End Region
#End Region
    End Class
End Namespace