﻿Option Explicit On
Option Strict On

Imports System.ComponentModel
Imports System.Globalization
Imports System.IO
Imports System.Runtime.InteropServices
Imports System.Security.Permissions
Imports System.Transactions

Imports Microsoft.Win32.SafeHandles
Imports MvpsOrg.Dotnet.Win32.NativeMethods
Imports MvpsOrg.Dotnet.Win32.SafeHandles

Namespace MvpsOrg.Dotnet.Win32.Transactions.IO
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class File
        Private Sub New()
            '
        End Sub

#Region "Copy and Move progress callback"
        Private NotInheritable Class CopyData
            Public ReadOnly CopyProgressCallback As New CopyProgressRoutine( _
                AddressOf CopyProgress _
            )

            Private m_CopyProgressCallback As CopyProgressCallback

            Public Sub New(ByVal CopyProgressCallback As CopyProgressCallback)
                m_CopyProgressCallback = CopyProgressCallback
            End Sub

            Private Function CopyProgress( _
                ByVal TotalFileSize As Int64, _
                ByVal TotalBytesTransferred As Int64, _
                ByVal StreamSize As Int64, _
                ByVal StreamBytesTransferred As Int64, _
                ByVal dwStreamNumber As Int32, _
                ByVal dwCallbackReason As CALLBACK_Constants, _
                ByVal hSourceFile As IntPtr, _
                ByVal hDestinationFile As IntPtr, _
                ByVal lpData As IntPtr _
            ) As CopyProgressRoutine_Result
                Return _
                    DirectCast( _
                        m_CopyProgressCallback( _
                            TotalFileSize, _
                            TotalBytesTransferred, _
                            StreamSize, _
                            StreamBytesTransferred, _
                            dwStreamNumber, _
                            DirectCast(dwCallbackReason, CallbackReason) _
                        ),  _
                        CopyProgressRoutine_Result _
                    )
            End Function
        End Class
#End Region

#Region "Copy"
#Region "Copy without progress"
#Region "Not transacted"
        Public Shared Sub Copy( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            Copy( _
                SourceFileName, _
                DestFileName, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub Copy( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As CopyFileOptions _
        )
            Dim Success As Boolean = _
                CopyFileEx( _
                    SourceFileName, _
                    DestFileName, _
                    Nothing, _
                    IntPtr.Zero, _
                    IntPtr.Zero, _
                    DirectCast(Options, COPY_FILE_Flags) _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CopyTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            CopyTransactedKtm( _
                SourceFileName, _
                DestFileName, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub CopyTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As CopyFileOptions _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                CopyFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As CopyFileOptions, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                CopyTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As CopyFileOptions _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Transaction As Transaction _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                CopyFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
             ByVal SourceFileName As String, _
             ByVal DestFileName As String, _
             ByVal Options As CopyFileOptions, _
             ByVal Transaction As Transaction _
         )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                CopyTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub CopyTransactedCore( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As CopyFileOptions, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                CopyFileTransacted( _
                    SourceFileName, _
                    DestFileName, _
                    Nothing, _
                    IntPtr.Zero, _
                    IntPtr.Zero, _
                    DirectCast(Options, COPY_FILE_Flags), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "Copy with progress"
#Region "Not transacted"
        Public Shared Sub Copy( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            Copy( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub Copy( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As CopyFileOptions _
        )
            Dim Data As New CopyData(ProgressCallback)
            Dim Success As Boolean = _
                CopyFileEx( _
                    SourceFileName, _
                    DestFileName, _
                    Data.CopyProgressCallback, _
                    IntPtr.Zero, _
                    IntPtr.Zero, _
                    DirectCast(Options, COPY_FILE_Flags) _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CopyTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            CopyTransactedKtm( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Transaction As Ktm.Transaction _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                CopyFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub CopyTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As CopyFileOptions _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
             ByVal SourceFileName As String, _
             ByVal DestFileName As String, _
             ByVal ProgressCallback As CopyProgressCallback, _
             ByVal Options As CopyFileOptions, _
             ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                CopyTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    ProgressCallback, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                CopyFileOptions.None _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Transaction As Transaction _
        )
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                CopyFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As CopyFileOptions _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            CopyTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CopyTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As CopyFileOptions, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                CopyTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    ProgressCallback, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub CopyTransactedCore( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As CopyFileOptions, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Data As New CopyData(ProgressCallback)
            Dim Success As Boolean = _
                CopyFileTransacted( _
                    SourceFileName, _
                    DestFileName, _
                    Data.CopyProgressCallback, _
                    IntPtr.Zero, _
                    IntPtr.Zero, _
                    DirectCast(Options, COPY_FILE_Flags), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region
#End Region

#Region "Move"
#Region "Move without progress"
#Region "Not transacted"
        Public Shared Sub Move( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            Move( _
                SourceFileName, _
                DestFileName, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub Move( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions _
        )
            Dim Success As Boolean = _
                MoveFileEx( _
                    SourceFileName, _
                    DestFileName, _
                    DirectCast(Options, MOVEFILE_Flags) _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub MoveTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            MoveTransactedKtm( _
                SourceFileName, _
                DestFileName, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                MoveFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub MoveTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                MoveTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Transaction As Transaction _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                MoveFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                MoveTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub MoveTransactedCore( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal Options As MoveFileOptions, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                MoveFileTransacted( _
                    SourceFileName, _
                    DestFileName, _
                    Nothing, _
                    IntPtr.Zero, _
                    DirectCast(Options, MOVEFILE_Flags), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "Move with progress"
#Region "Not transacted"
        Public Shared Sub Move( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            Move( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub Move( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions _
        )
            Dim Data As New CopyData(ProgressCallback)
            Dim Success As Boolean = _
                MoveFileWithProgress( _
                    SourceFileName, _
                    DestFileName, _
                    Data.CopyProgressCallback, _
                    Nothing, _
                    DirectCast(Options, MOVEFILE_Flags) _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub MoveTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            MoveTransactedKtm( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Transaction As Ktm.Transaction _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                MoveFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub MoveTransactedKtm( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                MoveTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    ProgressCallback, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                MoveFileOptions.None _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Transaction As Transaction _
        )
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                MoveFileOptions.None, _
                Transaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            MoveTransacted( _
                SourceFileName, _
                DestFileName, _
                ProgressCallback, _
                Options, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub MoveTransacted( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                MoveTransactedCore( _
                    SourceFileName, _
                    DestFileName, _
                    ProgressCallback, _
                    Options, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub MoveTransactedCore( _
            ByVal SourceFileName As String, _
            ByVal DestFileName As String, _
            ByVal ProgressCallback As CopyProgressCallback, _
            ByVal Options As MoveFileOptions, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Data As New CopyData(ProgressCallback)
            Dim Success As Boolean = _
                MoveFileTransacted( _
                    SourceFileName, _
                    DestFileName, _
                    Data.CopyProgressCallback, _
                    IntPtr.Zero, _
                    DirectCast(Options, MOVEFILE_Flags), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region
#End Region

#Region "Create"
#Region "Not transacted"
        Public Shared Function Create( _
            ByVal FileName As String _
        ) As FileStream
            Return Create(FileName, FileOptions.Normal)
        End Function

        Public Shared Function Create( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Dim FileHandle As SafeFileHandle = _
                CreateFile( _
                    FileName, _
                    DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                    FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                    New SECURITY_ATTRIBUTES(), _
                    CreationDisposition_Constants.CREATE_NEW, _
                    DirectCast(Options, FlagsAndAttributes_Flags), _
                    InvalidSafeFileHandle _
                )
            If FileHandle.IsInvalid Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return New FileStream(FileHandle, FileAccess.ReadWrite)
        End Function
#End Region

#Region "Transacted"
        Public Shared Function CreateTransactedKtm( _
            ByVal FileName As String _
        ) As FileStream
            Return _
                CreateTransactedKtm( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView _
                )
        End Function

        Public Shared Function CreateTransacted( _
             ByVal FileName As String, _
             ByVal Transaction As Transaction _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    FileOptions.None, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function CreateTransactedKtm( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Ktm.Transaction.Current _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction.Current _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal Transaction As Transaction _
        ) As FileStream
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function CreateTransactedKtm( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion _
        ) As FileStream
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    CreateTransactedCore( _
                        FileName, _
                        Options, _
                        MiniVersion, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion _
        ) As FileStream
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                CreateTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function CreateTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal Transaction As Transaction _
        ) As FileStream
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    CreateTransactedCore( _
                        FileName, _
                        Options, _
                        MiniVersion, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function CreateTransactedCore( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As FileStream
            Dim FileHandle As SafeFileHandle
            If MiniVersion = MiniVersion.OtherTransactionModifiesFile Then
                FileHandle = _
                    CreateFileTransacted( _
                        FileName, _
                        DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                        FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                        New SECURITY_ATTRIBUTES(), _
                        CreationDisposition_Constants.CREATE_NEW, _
                        DirectCast(Options, FlagsAndAttributes_Flags), _
                        InvalidSafeFileHandle, _
                        TransactionHandle, _
                        IntPtr.Zero, _
                        IntPtr.Zero _
                    )
            Else
                FileHandle = _
                    CreateFileTransacted( _
                        FileName, _
                        DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                        FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                        New SECURITY_ATTRIBUTES(), _
                        CreationDisposition_Constants.CREATE_NEW, _
                        DirectCast(Options, FlagsAndAttributes_Flags), _
                        InvalidSafeFileHandle, _
                        TransactionHandle, _
                        CType(MiniVersion, TXFS_MINIVERSION_Constants), _
                        IntPtr.Zero _
                    )
            End If
            If FileHandle.IsInvalid Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return New FileStream(FileHandle, FileAccess.ReadWrite)
        End Function
#End Region
#End Region

#Region "Open"
#Region "Not transacted"
        Public Shared Function Open( _
            ByVal FileName As String _
        ) As FileStream
            Return Open(FileName, FileOptions.Normal)
        End Function

        Public Shared Function Open( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Dim FileHandle As SafeFileHandle = _
                CreateFile( _
                    FileName, _
                    DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                    FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                    New SECURITY_ATTRIBUTES(), _
                    CreationDisposition_Constants.OPEN_EXISTING, _
                    DirectCast(Options, FlagsAndAttributes_Flags), _
                    InvalidSafeFileHandle _
                )
            If FileHandle.IsInvalid Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return New FileStream(FileHandle, FileAccess.ReadWrite)
        End Function
#End Region

#Region "Transacted"
        Public Shared Function OpenTransactedKtm( _
            ByVal FileName As String _
        ) As FileStream
            Return _
                OpenTransactedKtm( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    FileOptions.Normal, _
                    MiniVersion.DirtyView _
                )
        End Function

        Public Shared Function OpenTransacted( _
             ByVal FileName As String, _
             ByVal Transaction As Transaction _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    FileOptions.None, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function OpenTransactedKtm( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Ktm.Transaction.Current _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction.Current _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal Transaction As Transaction _
        ) As FileStream
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion.DirtyView, _
                    Transaction _
                )
        End Function

        Public Shared Function OpenTransactedKtm( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion _
        ) As FileStream
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileStream
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    OpenTransactedCore( _
                        FileName, _
                        Options, _
                        MiniVersion, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion _
        ) As FileStream
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                OpenTransacted( _
                    FileName, _
                    Options, _
                    MiniVersion, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function OpenTransacted( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal Transaction As Transaction _
        ) As FileStream
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    OpenTransactedCore( _
                        FileName, _
                        Options, _
                        MiniVersion, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function OpenTransactedCore( _
            ByVal FileName As String, _
            ByVal Options As FileOptions, _
            ByVal MiniVersion As MiniVersion, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As FileStream
            Dim FileHandle As SafeFileHandle
            If MiniVersion = MiniVersion.OtherTransactionModifiesFile Then
                FileHandle = _
                    CreateFileTransacted( _
                        FileName, _
                        DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                        FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                        New SECURITY_ATTRIBUTES(), _
                        CreationDisposition_Constants.OPEN_ALWAYS, _
                        DirectCast(Options, FlagsAndAttributes_Flags), _
                        InvalidSafeFileHandle, _
                        TransactionHandle, _
                        IntPtr.Zero, _
                        IntPtr.Zero _
                    )
            Else
                FileHandle = _
                    CreateFileTransacted( _
                        FileName, _
                        DesiredAccess_Flags.GENERIC_READ Or DesiredAccess_Flags.GENERIC_WRITE, _
                        FILE_SHARE_Flags.FILE_SHARE_READ Or FILE_SHARE_Flags.FILE_SHARE_WRITE, _
                        New SECURITY_ATTRIBUTES(), _
                        CreationDisposition_Constants.OPEN_ALWAYS, _
                        DirectCast(Options, FlagsAndAttributes_Flags), _
                        InvalidSafeFileHandle, _
                        TransactionHandle, _
                        CType(MiniVersion, TXFS_MINIVERSION_Constants), _
                        IntPtr.Zero _
                    )
            End If
            If FileHandle.IsInvalid Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return New FileStream(FileHandle, FileAccess.ReadWrite)
        End Function
#End Region
#End Region

#Region "Delete"
#Region "Not transacted"
        Public Shared Sub Delete( _
            ByVal FileName As String _
        )
            Dim Success As Boolean = DeleteFile(FileName)
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub DeleteTransactedKtm( _
            ByVal FileName As String _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            DeleteTransacted( _
                FileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub DeleteTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                DeleteTransactedCore(FileName, TransactionHandle)
            End Using
        End Sub

        Public Shared Sub DeleteTransacted( _
            ByVal FileName As String _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            DeleteTransacted( _
                FileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub DeleteTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                DeleteTransactedCore( _
                    FileName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub DeleteTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                DeleteFileTransacted( _
                    FileName, _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "CreateHardLink"
#Region "Not transacted"
        Public Shared Sub CreateHardLink( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String _
        )
            Dim Success As Boolean = _
                NativeMethods.CreateHardLink( _
                    NewFileName, _
                    ExistingFileName, _
                    New SECURITY_ATTRIBUTES() _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CreateHardLinkTransactedKtm( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CreateHardLinkTransacted( _
                NewFileName, _
                ExistingFileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateHardLinkTransacted( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Ktm.Transaction.Current.GetHandle()
                CreateHardLinkTransactedCore( _
                    NewFileName, _
                    ExistingFileName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CreateHardLinkTransacted( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            CreateHardLinkTransacted( _
                NewFileName, _
                ExistingFileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateHardLinkTransacted( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                CreateHardLinkTransactedCore( _
                    NewFileName, _
                    ExistingFileName, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub CreateHardLinkTransactedCore( _
            ByVal NewFileName As String, _
            ByVal ExistingFileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                NativeMethods.CreateHardLinkTransacted( _
                    NewFileName, _
                    ExistingFileName, _
                    New SECURITY_ATTRIBUTES(), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "CreateSymbolicLink"
#Region "Not transacted"
        Public Shared Sub CreateSymbolicLink( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetFileName As String _
        )
            Dim Success As Boolean = _
                NativeMethods.CreateSymbolicLink( _
                    SymbolicLinkFileName, _
                    TargetFileName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_FILE _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub CreateSymbolicLinkTransactedKtm( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetFileName As String _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            CreateSymbolicLinkTransacted( _
                SymbolicLinkFileName, _
                TargetFileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetFileName As String, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                CreateSymbolicLinkTransactedCore( _
                    SymbolicLinkFileName, _
                    TargetFileName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_FILE, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetFileName As String _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            CreateSymbolicLinkTransacted( _
                SymbolicLinkFileName, _
                TargetFileName, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub CreateSymbolicLinkTransacted( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetFileName As String, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                CreateSymbolicLinkTransactedCore( _
                    SymbolicLinkFileName, _
                    TargetFileName, _
                    SYMBOLIC_LINK_FLAG_Flags.SYMBOLIC_LINK_FLAG_FILE, _
                    TransactionHandle _
                )
            End Using
        End Sub

        ' Also used by 'Directory.CreateSymbolicLinkTransacted'.
        Friend Shared Sub CreateSymbolicLinkTransactedCore( _
            ByVal SymbolicLinkFileName As String, _
            ByVal TargetPath As String, _
            ByVal TargetType As SYMBOLIC_LINK_FLAG_Flags, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                NativeMethods.CreateSymbolicLinkTransacted( _
                    SymbolicLinkFileName, _
                    TargetPath, _
                    TargetType, _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "GetAttributes"
#Region "Not transacted"
        Public Shared Function GetAttributes( _
            ByVal FileName As String _
        ) As FileAttributes
            Dim Attributes As FILE_ATTRIBUTE_Flags = GetFileAttributes(FileName)
            If Attributes = INVALID_FILE_ATTRIBUTES Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return DirectCast(Attributes, FileAttributes)
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetAttributesKtm( _
            ByVal FileName As String _
        ) As FileAttributes
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetAttributes( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetAttributes( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As FileAttributes
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return GetAttributesCore( _
                    FileName, _
                    TransactionHandle _
                )
            End Using
        End Function

        Public Shared Function GetAttributesDtc( _
            ByVal FileName As String _
        ) As FileAttributes
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetAttributes( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetAttributes( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As FileAttributes
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return GetAttributesCore( _
                    FileName, _
                    TransactionHandle _
                )
            End Using
        End Function

        Private Shared Function GetAttributesCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As FileAttributes
            Return _
                DirectCast( _
                    GetFileAttributesCore( _
                        FileName, _
                        TransactionHandle _
                    ).dwFileAttributes,  _
                    FileAttributes _
                )
        End Function

        ' Used by 'GetAttributes', 'Get*Time*', and 'GetLengthTransacted'.
        Private Shared Function GetFileAttributesCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As WIN32_FILE_ATTRIBUTE_DATA
            Dim AttributeData As WIN32_FILE_ATTRIBUTE_DATA
            Dim Success As Boolean = _
                GetFileAttributesTransacted( _
                    FileName, _
                    GET_FILEEX_INFO_LEVELS.GetFileExInfoStandard, _
                    AttributeData, _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return AttributeData
        End Function
#End Region
#End Region

#Region "SetAttributes"
#Region "Not transacted"
        Public Shared Sub SetAttributes( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes _
        )
            Dim Success As Boolean = _
                SetFileAttributes( _
                    FileName, _
                    DirectCast(NewAttributes, FILE_ATTRIBUTE_Flags) _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region

#Region "Transacted"
        Public Shared Sub SetAttributesKtm( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes _
        )
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            SetAttributes( _
                FileName, _
                NewAttributes, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub SetAttributes( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes, _
            ByVal Transaction As Ktm.Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                SetAttributesCore( _
                    FileName, _
                    NewAttributes, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Public Shared Sub SetAttributesDtc( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes _
        )
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            SetAttributes( _
                FileName, _
                NewAttributes, _
                CurrentTransaction _
            )
        End Sub

        Public Shared Sub SetAttributes( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes, _
            ByVal Transaction As Transaction _
        )
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                SetAttributesCore( _
                    FileName, _
                    NewAttributes, _
                    TransactionHandle _
                )
            End Using
        End Sub

        Private Shared Sub SetAttributesCore( _
            ByVal FileName As String, _
            ByVal NewAttributes As FileAttributes, _
            ByVal TransactionHandle As SafeTransactionHandle _
        )
            Dim Success As Boolean = _
                SetFileAttributesTransacted( _
                    FileName, _
                    DirectCast(NewAttributes, FILE_ATTRIBUTE_Flags), _
                    TransactionHandle _
                )
            If Not Success Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
        End Sub
#End Region
#End Region

#Region "GetCreationTimeTransacted"
        Public Shared Function GetCreationTimeTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetCreationTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCreationTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetCreationTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetCreationTimeTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetCreationTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCreationTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetCreationTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetCreationTimeTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftCreationTime
            Return _
                Date.FromFileTime( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetLastAccessTimeTransacted"
        Public Shared Function GetLastAccessTimeTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLastAccessTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastAccessTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLastAccessTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLastAccessTimeTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLastAccessTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastAccessTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLastAccessTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLastAccessTimeTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftLastAccessTime
            Return _
                Date.FromFileTime( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetLastWriteTimeTransacted"
        Public Shared Function GetLastWriteTimeTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLastWriteTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastWriteTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLastWriteTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLastWriteTimeTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLastWriteTimeTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastWriteTimeTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLastWriteTimeTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLastWriteTimeTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftLastWriteTime
            Return _
                Date.FromFileTime( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetCreationTimeUtcTransacted"
        Public Shared Function GetCreationTimeUtcTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetCreationTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCreationTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetCreationTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetCreationTimeUtcTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetCreationTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCreationTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetCreationTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetCreationTimeUtcTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftCreationTime
            Return _
                Date.FromFileTimeUtc( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetLastAccessTimeUtcTransacted"
        Public Shared Function GetLastAccessTimeUtcTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLastAccessTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastAccessTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLastAccessTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLastAccessTimeUtcTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLastAccessTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastAccessTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLastAccessTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLastAccessTimeUtcTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftLastAccessTime
            Return _
                Date.FromFileTimeUtc( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetLastWriteTimeUtcTransacted"
        Public Shared Function GetLastWriteTimeUtcTransactedKtm( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLastWriteTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastWriteTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLastWriteTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLastWriteTimeUtcTransacted( _
            ByVal FileName As String _
        ) As Date
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLastWriteTimeUtcTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLastWriteTimeUtcTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Date
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLastWriteTimeUtcTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLastWriteTimeUtcTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Date
            Dim FileTime As ComTypes.FILETIME = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                ).ftLastWriteTime
            Return _
                Date.FromFileTimeUtc( _
                    WordConverter.MakeLong( _
                        FileTime.dwLowDateTime, _
                        FileTime.dwHighDateTime _
                    ) _
                )
        End Function
#End Region

#Region "GetLengthTransacted"
        Public Shared Function GetLengthTransactedKtm( _
            ByVal FileName As String _
        ) As Long
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLengthTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLengthTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Long
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLengthTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLengthTransacted( _
            ByVal FileName As String _
        ) As Long
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLengthTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLengthTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Long
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLengthTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLengthTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Long
            Dim FileAttributes As WIN32_FILE_ATTRIBUTE_DATA = _
                GetFileAttributesCore( _
                    FileName, _
                    TransactionHandle _
                )
            Return _
                WordConverter.MakeLong( _
                    FileAttributes.nFileSizeLow, _
                    FileAttributes.nFileSizeHigh _
                )
        End Function
#End Region

#Region "GetCompressedLength"
#Region "Not transacted"
        Public Shared Function GetCompressedLength( _
            ByVal FileName As String _
        ) As Long
            Dim LoFileSize, HiFileSize As Integer
            LoFileSize = _
                GetCompressedFileSize( _
                    FileName, _
                    HiFileSize _
                )
            If _
                (HiFileSize = 0 AndAlso LoFileSize = INVALID_FILE_SIZE) _
                OrElse _
                (HiFileSize <> 0 AndAlso LoFileSize = INVALID_FILE_ATTRIBUTES AndAlso Marshal.GetLastWin32Error() <> NO_ERROR) _
            Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return WordConverter.MakeLong(LoFileSize, HiFileSize)
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetCompressedLengthKtm( _
            ByVal FileName As String _
        ) As Long
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetCompressedLength( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCompressedLength( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As Long
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetCompressedLengthCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetCompressedLengthDtc( _
            ByVal FileName As String _
        ) As Long
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetCompressedLength( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetCompressedLength( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As Long
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetCompressedLengthCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetCompressedLengthCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As Long
            Dim LoFileSize, HiFileSize As Integer
            LoFileSize = _
                GetCompressedFileSizeTransacted( _
                    FileName, _
                    HiFileSize, _
                    TransactionHandle _
                )
            If _
                (HiFileSize = 0 AndAlso LoFileSize = INVALID_FILE_SIZE) _
                OrElse _
                (HiFileSize <> 0 AndAlso LoFileSize = INVALID_FILE_ATTRIBUTES AndAlso Marshal.GetLastWin32Error() <> NO_ERROR) _
            Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            Return WordConverter.MakeLong(LoFileSize, HiFileSize)
        End Function
#End Region
#End Region
    End Class
End Namespace