﻿Option Explicit On
Option Strict On

Imports System.ComponentModel
Imports System.Globalization
Imports System.IO
Imports System.Security.Permissions
Imports System.Transactions

Imports MvpsOrg.Dotnet.Win32.NativeMethods
Imports MvpsOrg.Dotnet.Win32.SafeHandles

Namespace MvpsOrg.Dotnet.Win32.Transactions.IO
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class Path
        Private Sub New()
            '
        End Sub

#Region "GetLongPath"
#Region "Not transacted"
        Public Shared Function GetLongPath( _
            ByVal Path As String _
        ) As String
            Const BufferLength As Integer = MAX_PATH
            Dim LongPath As String = Space(BufferLength)
            Dim LongPathLength As Integer = _
                GetLongPathName( _
                    Path, _
                    LongPath, _
                    LongPath.Length _
                )
            If Not CBool(LongPathLength) Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            If LongPathLength <= BufferLength Then
                Return Left(LongPath, LongPathLength)
            End If
            LongPath = Space(LongPathLength)
            LongPathLength = _
                GetLongPathName( _
                    Path, _
                    LongPath, _
                    LongPath.Length _
                )

            ' No checking for errors again.
            Return LongPath
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetLongPathTransactedKtm( _
            ByVal Path As String _
        ) As String
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetLongPathTransacted( _
                    Path, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLongPathTransacted( _
            ByVal Path As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As String
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetLongPathTransactedCore( _
                        Path, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetLongPathTransacted( _
            ByVal Path As String _
        ) As String
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetLongPathTransacted( _
                    Path, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetLongPathTransacted( _
            ByVal Path As String, _
            ByVal Transaction As Transaction _
        ) As String
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetLongPathTransactedCore( _
                        Path, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetLongPathTransactedCore( _
            ByVal Path As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As String
            Const BufferLength As Integer = MAX_PATH
            Dim LongPath As String = Space(BufferLength)
            Dim LongPathLength As Integer = _
                GetLongPathNameTransacted( _
                    Path, _
                    LongPath, _
                    LongPath.Length, _
                    TransactionHandle _
                )
            If Not CBool(LongPathLength) Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            If LongPathLength <= BufferLength Then
                Return Left(LongPath, LongPathLength)
            End If
            LongPath = Space(LongPathLength)
            LongPathLength = _
                GetLongPathNameTransacted( _
                    Path, _
                    LongPath, _
                    LongPath.Length, _
                    TransactionHandle _
                )

            ' No checking for errors again.
            Return LongPath
        End Function
#End Region
#End Region

#Region "GetFullPath"
#Region "Not transacted"
        Public Shared Function GetFullPath( _
            ByVal FileName As String _
        ) As String
            Const BufferLength As Integer = MAX_PATH
            Dim LongPath As String = Space(BufferLength)
            Dim LongPathLength As Integer = _
                GetFullPathName( _
                    FileName, _
                    LongPath.Length, _
                    LongPath, _
                    IntPtr.Zero _
                )
            If Not CBool(LongPathLength) Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            If LongPathLength <= BufferLength Then
                Return Left(LongPath, LongPathLength)
            End If
            LongPath = Space(LongPathLength)
            LongPathLength = _
                GetFullPathName( _
                    FileName, _
                    LongPath.Length, _
                    LongPath, _
                    IntPtr.Zero _
                )

            ' No checking for errors again.
            Return LongPath
        End Function
#End Region

#Region "Transacted"
        Public Shared Function GetFullPathTransactedKtm( _
            ByVal FileName As String _
        ) As String
            Dim CurrentTransaction As Ktm.Transaction = Ktm.Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentKtmTransaction)
            End If
            Return _
                GetFullPathTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetFullPathTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Ktm.Transaction _
        ) As String
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = Transaction.GetHandle()
                Return _
                    GetFullPathTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Public Shared Function GetFullPathTransacted( _
            ByVal FileName As String _
        ) As String
            Dim CurrentTransaction As Transaction = Transaction.Current
            If CurrentTransaction Is Nothing Then
                Throw New InvalidOperationException(My.Resources.ExceptionNoCurrentDtcTransaction)
            End If
            Return _
                GetFullPathTransacted( _
                    FileName, _
                    CurrentTransaction _
                )
        End Function

        Public Shared Function GetFullPathTransacted( _
            ByVal FileName As String, _
            ByVal Transaction As Transaction _
        ) As String
            If Transaction Is Nothing Then
                Throw New ArgumentException(My.Resources.ExceptionParameterShouldHaveValidValueButIsNothing, "Transaction")
            End If
            Using TransactionHandle As SafeTransactionHandle = SafeTransactionHandle.GetHandleFromDtcTransaction(Transaction)
                Return _
                    GetFullPathTransactedCore( _
                        FileName, _
                        TransactionHandle _
                    )
            End Using
        End Function

        Private Shared Function GetFullPathTransactedCore( _
            ByVal FileName As String, _
            ByVal TransactionHandle As SafeTransactionHandle _
        ) As String
            Const BufferLength As Integer = MAX_PATH
            Dim LongPath As String = Space(BufferLength)
            Dim LongPathLength As Integer = _
                GetFullPathNameTransacted( _
                    FileName, _
                    LongPath.Length, _
                    LongPath, _
                    IntPtr.Zero, _
                    TransactionHandle _
                )
            If Not CBool(LongPathLength) Then
                Dim ex As New Win32Exception()
                Throw New IOException(ex.Message, ex)
            End If
            If LongPathLength <= BufferLength Then
                Return Left(LongPath, LongPathLength)
            End If
            LongPath = Space(LongPathLength)
            LongPathLength = _
                GetFullPathNameTransacted( _
                    FileName, _
                    LongPath.Length, _
                    LongPath, _
                    IntPtr.Zero, _
                    TransactionHandle _
                )

            ' No checking for errors again.
            Return LongPath
        End Function
#End Region
#End Region
    End Class
End Namespace