﻿Option Explicit On
Option Strict On

Imports System.ComponentModel
Imports System.Diagnostics.CodeAnalysis
Imports System.Security.Permissions
Imports System.Threading

Imports MvpsOrg.Dotnet.Win32.NativeMethods
Imports MvpsOrg.Dotnet.Win32.SafeHandles

Namespace MvpsOrg.Dotnet.Win32.Transactions.Ktm
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class Transaction
        Implements IDisposable

#Region "Shared members"
        Public Shared ReadOnly TimeoutInfinite As New TimeSpan(0)

        Public Shared Property Current() As Transaction
            Get
                Return TransactionScope.CurrentAmbientTransaction
            End Get
            Set(ByVal Value As Transaction)
                TransactionScope.CurrentAmbientTransaction = Value
            End Set
        End Property
#End Region

        Private m_TransactionHandle As SafeTransactionHandle
        Private m_State As TransactionStatus = TransactionStatus.Active

        Public Sub New()
            Me.New(TimeoutInfinite, Nothing)
        End Sub

        Public Sub New( _
            ByVal Timeout As TimeSpan _
        )
            Me.New(Timeout, Nothing)
        End Sub

        Public Sub New( _
            ByVal Timeout As TimeSpan, _
            ByVal Description As String _
        )
            m_TransactionHandle = _
                CreateTransaction( _
                    Nothing, _
                    IntPtr.Zero, _
                    0, _
                    0, _
                    0, _
                    CInt(Timeout.TotalMilliseconds), _
                    Description _
                )
            If m_TransactionHandle.IsInvalid Then
                Throw New Win32Exception()
            End If
        End Sub

        Public ReadOnly Property State() As TransactionStatus
            Get
                Return m_State
            End Get
        End Property

        Public Sub Commit()
            If m_State <> TransactionStatus.Active Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionAlreadyCommittedOrRolledBack)
            End If
            If Not CommitTransaction(m_TransactionHandle) Then
                Throw New Win32Exception()
            End If
            m_State = TransactionStatus.Committed
        End Sub

        Public Sub BeginCommit()
            If m_State <> TransactionStatus.Active Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionAlreadyCommittedOrRolledBack)
            End If
            If Not CommitTransactionAsync(m_TransactionHandle) Then
                Throw New Win32Exception()
            End If
            m_State = TransactionStatus.Committed
        End Sub

        Public Sub Rollback()
            If m_State <> TransactionStatus.Active Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionAlreadyCommittedOrRolledBack)
            End If
            If Not RollbackTransaction(m_TransactionHandle) Then
                Throw New Win32Exception()
            End If
            m_State = TransactionStatus.Aborted
        End Sub

        Public Sub BeginRollback()
            If m_State <> TransactionStatus.Active Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionAlreadyCommittedOrRolledBack)
            End If
            If Not RollbackTransactionAsync(m_TransactionHandle) Then
                Throw New Win32Exception()
            End If
            m_State = TransactionStatus.Aborted
        End Sub

        <SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")> _
        Public Function GetHandle() As SafeTransactionHandle
            Dim TransactionHandle As SafeTransactionHandle = _
                OpenTransaction( _
                    TRANSACTION_ACCESS_Flags.TRANSACTION_ALL_ACCESS, _
                    Me.GetId() _
                )
            If TransactionHandle.IsInvalid Then
                Throw New Win32Exception()
            End If
            Return TransactionHandle
        End Function

        Public Function Id() As System.Guid
            Dim TransactionId As GUID = Me.GetId()
            Return _
                New System.Guid( _
                    TransactionId.Data1, _
                    TransactionId.Data2, _
                    TransactionId.Data3, _
                    TransactionId.Data4 _
                )
        End Function

        Private Function GetId() As GUID
            Dim TransactionId As New GUID()
            If Not GetTransactionId(m_TransactionHandle, TransactionId) Then
                Throw New Win32Exception()
            End If
            Return TransactionId
        End Function

#Region "IDisposable"
        Private m_IsDisposed As Boolean

        Private Sub Dispose(ByVal Disposing As Boolean)
            If Not m_IsDisposed Then
                If Disposing Then

                    ' TODO: Release managed resources.
                    If Not m_TransactionHandle.IsInvalid Then
                        m_TransactionHandle.Dispose()
                    End If
                End If

                ' TODO: Release unmanaged resources.
            End If
            m_IsDisposed = True
        End Sub

        Public Sub Dispose() Implements IDisposable.Dispose
            Me.Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region
    End Class
End Namespace