﻿Option Explicit On
Option Strict On

Imports System.Security.Permissions
Imports System.Threading

Namespace MvpsOrg.Dotnet.Win32.Transactions.Ktm
    <SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode:=True)> _
    Public NotInheritable Class TransactionScope
        Implements IDisposable

#Region "Shared members"
        <ThreadStatic()> _
        Private Shared m_ThreadTransactionContext As New ThreadTransactionContext()

        ' The current ambient transaction (either set by a scope or by
        ' assigning the transaction using 'Transaction.Current'.
        Friend Shared Property CurrentAmbientTransaction() As Transaction
            Get
                Return m_ThreadTransactionContext.CurrentTransaction
            End Get
            Set(ByVal Value As Transaction)
                m_ThreadTransactionContext.CurrentTransaction = Value
            End Set
        End Property
#End Region

#Region "Nested types"
        Private Class ThreadTransactionContext
            Private m_ScopeStack As New Stack(Of ScopeData)()
            Private m_CurrentTransaction As Transaction

            Public Sub Push(ByVal ScopeData As ScopeData)
                m_ScopeStack.Push(ScopeData)
                m_CurrentTransaction = ScopeData.ScopeTransaction
            End Sub

            Public Sub Pop()
                m_ScopeStack.Pop()
                If m_ScopeStack.Count > 0 Then
                    m_CurrentTransaction = m_ScopeStack.Peek().ScopeTransaction
                End If
            End Sub

            Public Function Peek() As ScopeData
                Return m_ScopeStack.Peek
            End Function

            Public ReadOnly Property Count() As Integer
                Get
                    Return m_ScopeStack.Count
                End Get
            End Property

            Public Property CurrentTransaction() As Transaction
                Get
                    Return m_CurrentTransaction
                End Get
                Set(ByVal Value As Transaction)
                    m_CurrentTransaction = Value
                End Set
            End Property
        End Class

        Private Class ScopeData
            Public Sub New( _
                ByVal ScopeTransaction As Transaction, _
                ByVal ScopeThread As Thread _
            )
                Me.ScopeTransaction = ScopeTransaction
                Me.ScopeThread = ScopeThread
            End Sub

            Public ScopeTransaction As Transaction
            Public ScopeThread As Thread
        End Class
#End Region

        ' The transaction associated with the scope ('Nothing' if
        ' 'ScopeOption.Suppress' is used).
        Private m_Transaction As Transaction

        ' The scope has been marked as complete and the transaction is
        ' committed if it is owned by the scope.
        Private m_Complete As Boolean

        ' The transaction is owned by the scope and is committed and
        ' destroyed by the scope.
        Private m_ScopeOwnsTransaction As Boolean

        Public Sub New()
            Me.New( _
                TransactionScopeOption.Required, _
                Transaction.TimeoutInfinite _
            )
        End Sub

        Public Sub New( _
            ByVal ScopeOption As TransactionScopeOption _
        )
            Me.New( _
                ScopeOption, _
                Transaction.TimeoutInfinite _
            )
        End Sub

        Public Sub New( _
            ByVal Timeout As TimeSpan _
        )
            Me.New( _
                TransactionScopeOption.RequiredNew, _
                Timeout _
            )
        End Sub

        Public Sub New( _
            ByVal ScopeOption As TransactionScopeOption, _
            ByVal Timeout As TimeSpan _
        )
            Select Case ScopeOption
                Case TransactionScopeOption.Required
                    SyncLock m_ThreadTransactionContext
                        If m_ThreadTransactionContext.Count = 0 Then
                            m_Transaction = New Transaction(Timeout)
                            m_ThreadTransactionContext.Push( _
                                New ScopeData( _
                                    m_Transaction, _
                                    Thread.CurrentThread _
                                ) _
                            )
                            m_ScopeOwnsTransaction = True
                        End If
                    End SyncLock
                Case TransactionScopeOption.RequiredNew
                    SyncLock m_ThreadTransactionContext
                        m_Transaction = New Transaction(Timeout)
                        m_ThreadTransactionContext.Push( _
                            New ScopeData( _
                                m_Transaction, _
                                Thread.CurrentThread _
                            ) _
                        )
                        m_ScopeOwnsTransaction = True
                    End SyncLock
                Case TransactionScopeOption.Suppress
                    SyncLock m_ThreadTransactionContext
                        m_ThreadTransactionContext.CurrentTransaction = Nothing
                    End SyncLock
            End Select
        End Sub

        Public Sub New( _
            ByVal TransactionToUse As Transaction _
        )
            SyncLock m_ThreadTransactionContext
                m_Transaction = TransactionToUse
                m_ThreadTransactionContext.Push( _
                    New ScopeData( _
                        m_Transaction, _
                        Thread.CurrentThread _
                    ) _
                )
            End SyncLock
        End Sub

        Public Sub Complete()
            If m_Complete Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionScopeAlreadyMarkedAsComplete)
            End If
            m_Complete = True
        End Sub

#Region "IDisposable"
        Private m_IsDisposed As Boolean

        Private Sub Dispose(ByVal Disposing As Boolean)

            ' 'Dispose' must not be called from another thread than those
            ' the instance has been created on.
            If Thread.CurrentThread IsNot m_ThreadTransactionContext.Peek().ScopeThread Then
                Throw New InvalidOperationException(My.Resources.ExceptionTransactionScopeMustNotBeDisposedFromAnotherThreadThanItHasBeenCreatedOn)
            End If

            If Not m_IsDisposed Then
                If Disposing Then

                    ' TODO: Release managed resources.
                    If m_ScopeOwnsTransaction Then
                        Try
                            If m_ThreadTransactionContext.CurrentTransaction IsNot m_Transaction Then
                                Throw New InvalidOperationException(My.Resources.ExceptionTransactionCurrentChangedInTransactionScope)
                            End If
                            Using m_Transaction
                                If m_Complete Then
                                    m_Transaction.Commit()
                                Else
                                    m_Transaction.Rollback()
                                End If
                            End Using
                        Finally
                            SyncLock m_ThreadTransactionContext
                                m_ThreadTransactionContext.Pop()
                            End SyncLock
                        End Try
                    Else
                        If Not m_Complete Then
                            Using m_Transaction
                                m_Transaction.Rollback()
                            End Using
                        End If
                    End If
                End If

                ' TODO: Release unmanaged resources.
            End If
            m_IsDisposed = True
        End Sub

        Public Sub Dispose() Implements IDisposable.Dispose
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region
    End Class
End Namespace